/*
This is part of the audio CD player library
Copyright (C)1998 Tony Arcieri

All changed initialed (MO) made by Mike Oliphant (oliphant@ling.ed.ac.uk)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.


 o fixed problem with snprintf() not liking src & dest to be the same (MO)
 o fixed problems handling albums without titles (MO)
 o switched code to use http instead of the (now deprecated) cddb

*/

#include <cdaudio.h>
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <pwd.h>
#include <errno.h>
#include <unistd.h>
#include <cddb.h>

/* Static function prototypes */
static int cddb_sum(int val);
static int cddb_conf_data_mod(char *var, char *value, struct cddb_conf *conf);
static int cddb_conf_process_line(char *line, struct cddb_conf *conf);
static void cddb_skip_http(int sock);
static int cddb_read_line(int sock, char *inbuffer, int len);
static int cddb_get_server_force(struct cddb_server *server);
static void cddb_make_hello(struct cddb_hello *hello,char *hellobuf);
static void cddb_make_request(struct cddb_server *server,
							  struct cddb_hello *hello,
							  char *cmd,char *outbuf);
static int cddb_data_mod(char *var, char *value,
						 struct __unprocessed_disc_data *data);
static int cddb_process_line(char *line, struct __unprocessed_disc_data *data);
static int cddb_sites_process_line(char *line, struct cddb_hostelement *host);
static int cddb_sites(struct cddb_server *server,struct cddb_serverlist *list);

/* CDDB sum function... got me, they invented it */
static int
cddb_sum(int val)
{
   char *bufptr, buf[16];
   int ret = 0;
   
   snprintf(buf, 16, "%lu", val);
   for(bufptr = buf; *bufptr != '\0'; bufptr++)
     ret += (*bufptr - '0');
   
   return ret;
}

/* Produce CDDB ID for CD currently in CD-ROM */
unsigned long 
cddb_discid(int cd_desc)
{
  int index, tracksum = 0, discid;
  struct disc_info disc;
  
  cd_stat(cd_desc, &disc);
  
  for(index = 0; index < disc.disc_totaltracks; index++)
	tracksum += cddb_sum(disc.track[index].track_pos.minutes * 60 +
						 disc.track[index].track_pos.seconds);
  
  discid = (disc.disc_length.minutes * 60 + disc.disc_length.seconds) -
	(disc.track[0].track_pos.minutes * 60 + disc.track[0].track_pos.seconds);
  
  return (tracksum % 0xFF) << 24 | discid << 8 | disc.disc_totaltracks;
}

/* Modify a cddb_conf structure based upon a CDDB configuration command */
static int
cddb_conf_data_mod(char *var, char *value, struct cddb_conf *conf)
{   
  if(value == NULL)
	value = "";
  
  if(strcasecmp(var, "cddb") == 0) {
	if(strncasecmp(value, "on", 2) == 0)
	  conf->enable = 1;
	else
	  conf->enable = 0;
  } else if(strcasecmp(var, "server") == 0) {
	if(conf->hostlen >= CDDB_MAX_SERVERS)
	  return 0;
	if(strchr(value, ':') == NULL) {
	  strncpy(conf->server[conf->hostlen].server_name, value, 256);
	  conf->server[conf->hostlen++].server_port = CDDB_DEFAULT_PORT;
	} else {
	  strncpy(conf->server[conf->hostlen].server_name, strtok(value, ":"),
			  256);
	 conf->server[conf->hostlen++].server_port = strtol(strtok(NULL, ":"),
														NULL, 10);
	}
  }
  
  return 0;
}

/* Process a line in a CDDB configuration file */
static int
cddb_conf_process_line(char *line, struct cddb_conf *conf)
{
  char *var, *value;
  
  if(strchr(line, ' ') == NULL)
	return 0;
  
  line[strlen(line) - 1] = '\0';
  
  var = strtok(line, " ");
  if(var == NULL)
	return 0;
  value = strtok(NULL, " ");
  cddb_conf_data_mod(var, value, conf);
  
  return 0;
}

/* Read ~/.cddbrc and /etc/cddb.conf */
static int
cddb_read_conf(struct cddb_conf *conf)
{
  FILE *cddbconf;
  int index, rconf;
  char inbuffer[256];
  char localconfpath[256];
  struct stat st;
  
  conf->hostlen = 0;
  conf->enable = 1;         /* CDDB support defaults to on */
  rconf = 0;
  
  /* The user's server list takes priority over the global server list */
  snprintf(localconfpath, 256, "%s/.cddbrc", getenv("HOME"));
  if(stat(localconfpath, &st) == 0) {
	cddbconf = fopen(localconfpath, "r");
	
	while(!feof(cddbconf)) {
	  fgets(inbuffer, 256, cddbconf);
	  inbuffer[255] = '\0';
      
	  /* Remove comments */
      
	  for(index = 0; index < strlen(inbuffer); index++)
		if(inbuffer[index] == '#') {
	      inbuffer[index] = '\0';
	      break;
		}
      
	  cddb_conf_process_line(inbuffer, conf);
	}
	
	fclose(cddbconf);
	rconf = 1;
  }
  
  if(stat(CDDB_CONF_PATH, &st) == 0) {
	cddbconf = fopen(CDDB_CONF_PATH, "r");
	
	while(!feof(cddbconf)) {
	  fgets(inbuffer, 256, cddbconf);
	  inbuffer[255] = '\0';
      
	  /* Remove comments */
      
	  for(index = 0; index < strlen(inbuffer); index++)
		if(inbuffer[index] == '#') {
	      inbuffer[index] = '\0';
	      break;
		}
	  
	  /* User configuration takes priority over global */
	  if(strncasecmp(inbuffer, "cddb", 4) != 0 || rconf != 1)
		cddb_conf_process_line(inbuffer, conf);
	}
	
	fclose(cddbconf);
	rconf = 1;
  }
  
  if(rconf == 0)
	return -1;
  
  return 0;
}

/* Convert numerical genre to text */
char
*cddb_genre(int genre)
{
  switch(genre) {
  case CDDB_BLUES:
	return "blues";
  case CDDB_CLASSICAL:
	return "classical";
  case CDDB_COUNTRY:
	return "country";
  case CDDB_DATA:
	return "data";
  case CDDB_FOLK:
	return "folk";
  case CDDB_JAZZ:
	return "jazz";
  case CDDB_MISC:
	return "misc";
  case CDDB_NEWAGE:
	return "newage";
  case CDDB_REGGAE:
	return "reggae";
  case CDDB_ROCK:
	return "rock";
  case CDDB_SOUNDTRACK:
	return "soundtrack";
  }
  
  return "unknown";
}

/* Convert genre from text form into an integer value */
int
cddb_genre_value(char *genre)
{
  if(strcmp(genre, "blues") == 0)
	return CDDB_BLUES;
  else if(strcmp(genre, "classical") == 0)
	return CDDB_CLASSICAL;
  else if(strcmp(genre, "country") == 0)
	return CDDB_COUNTRY;
  else if(strcmp(genre, "data") == 0)
	return CDDB_DATA;
  else if(strcmp(genre, "folk") == 0)
	return CDDB_FOLK;
  else if(strcmp(genre, "jazz") == 0)
	return CDDB_JAZZ;
  else if(strcmp(genre, "misc") == 0)
	return CDDB_MISC;
  else if(strcmp(genre, "newage") == 0)
	return CDDB_NEWAGE;
  else if(strcmp(genre, "reggae") == 0)
	return CDDB_REGGAE;
  else if(strcmp(genre, "rock") == 0)
	return CDDB_ROCK;
  else if(strcmp(genre, "soundtrack") == 0)
	return CDDB_SOUNDTRACK;
  else
	return CDDB_UNKNOWN;
}

/* Connect to a CDDB server */
static int
cddb_connect_server(struct cddb_server *server)
{
  int sock;
  struct sockaddr_in sin;
  struct hostent *host;
  char *sname;
  
  sin.sin_family = AF_INET;

  if(server->server_use_proxy)
	sin.sin_port=htons(server->server_proxy_port);
  else
	sin.sin_port = htons(server->server_port);
  
  if(server->server_use_proxy)
	sname=server->server_proxy;
  else
	sname=server->server_name;
  
  if((sin.sin_addr.s_addr = inet_addr(sname)) == INADDR_NONE) {
	if((host = gethostbyname(sname)) == NULL)
	  return -1;
	bcopy(host->h_addr, &sin.sin_addr, host->h_length);
  }

  if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return -1;

  if(connect(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0)
	return -1;

  return sock;
}

/* Skip http header */

static void cddb_skip_http(int sock)
{
  char inchar;
  int len;

  do {
	len=0;

	do {
	  read(sock,&inchar,1);
	  len++;
	}
	while(inchar!='\n');
  }
  while(len>2);
}

/* Read a single line */
static int
cddb_read_line(int sock, char *inbuffer, int len)
{
  int index;
  char inchar;
  
  for(index = 0; index < len; index++) {
	read(sock, &inchar, 1);

	if(inchar == '\n') {
	  inbuffer[index] = '\0';
	  if(inbuffer[0] == '.')
		return 1;
	  return 0;
	}
	inbuffer[index] = inchar;
  }
  
  return index;
}

/* Pull entry from CDDB configuration files */
int
cddb_get_server(struct cddb_server *server)
{
  struct cddb_conf conf;
   
  /* If we can't find any server names, use the defaults */
  if(cddb_read_conf(&conf) < 0 || conf.hostlen == 0) {
	strncpy(server->server_name, CDDB_DEFAULT_HOSTNAME, 256);
	server->server_port = CDDB_DEFAULT_PORT;
	server->server_use_proxy=0;
	
	return 0;
  }
  
  if(conf.enable == 0)
	return -1;
  
  if(server->server_index >= conf.hostlen)
	return -1;
  
  strncpy(server->server_name, conf.server[server->server_index].server_name,
		  256);
  server->server_port = conf.server[server->server_index].server_port;
  server->server_use_proxy=0;
   
  return 0;
}

/* Get an entry from the server list whether CDDB is on or off */
static int
cddb_get_server_force(struct cddb_server *server)
{
  struct cddb_conf conf;

  /* If we can't find any server names, use the defaults */
  if(cddb_read_conf(&conf) < 0 || conf.hostlen == 0) {
	strncpy(server->server_name, CDDB_DEFAULT_HOSTNAME, 256);
	server->server_port = CDDB_DEFAULT_PORT;
	
	return 0;
  }
  
  if(server->server_index >= conf.hostlen)
	return -1;
  
  strncpy(server->server_name, conf.server[server->server_index].server_name,
		  256);
  server->server_port = conf.server[server->server_index].server_port;
   
  return 0;
}

/* Make a 'hello' string from a cddb_hello structure */

static void cddb_make_hello(struct cddb_hello *hello,char *hellobuf)
{
   struct passwd *pw;
   char outbuffer[256];
   char hostname[256];
   
   pw=getpwuid(getuid());
   gethostname(hostname,256);

   snprintf(hellobuf,256,"&hello=%s+%s+%s+%s",pw->pw_name,hostname,
			hello->hello_program,hello->hello_version);
}

/* Make a CDDB http request string */

static void cddb_make_request(struct cddb_server *server,
							  struct cddb_hello *hello,
							  char *cmd,char *outbuf)
{
  char hellobuf[256];  

  cddb_make_hello(hello,hellobuf);

  if(server->server_use_proxy)
	snprintf(outbuf,256,
			 "GET http://%s/~cddb/cddb.cgi?cmd=%s%s&proto=%s HTTP/1.0\n\n",
			 server->server_name,cmd,hellobuf,CDDB_PROTOCOL_LEVEL);
  else
	snprintf(outbuf,256,"GET /~cddb/cddb.cgi?cmd=%s%s&proto=%s HTTP/1.0\n\n",
			 cmd,hellobuf,CDDB_PROTOCOL_LEVEL);
}

/* Query the CDDB for the CD currently in the CD-ROM, and find the ID of the
   CD (which may or may not be the one generated) and what section it is
   under (genre) */
int
cddb_query(int cd_desc,struct cddb_server *server,struct cddb_hello *hello,
		   struct cddb_query *query)
{
  int socket;
  int index;
  struct disc_info disc;
  char offsetbuffer[256], outbuffer[256],
	inbuffer[256], tmpbuffer[256], procbuffer[256];
  char *tmptr, idtext[16];
  char *token;

  socket=cddb_connect_server(server);
  if(socket==-1)
	return -1;
  
  query->query_matches = 0;
  cd_stat(cd_desc, &disc);
  
  snprintf(offsetbuffer, 256, "%d", disc.disc_totaltracks);
  
  for(index = 0; index < disc.disc_totaltracks; index++)
	sprintf(offsetbuffer, "%s+%d",offsetbuffer,
			disc.track[index].track_start);

  snprintf(tmpbuffer,256,"cddb+query+%08lx+%s+%d",cddb_discid(cd_desc),
		   offsetbuffer, disc.disc_length.minutes * 60 +
		   disc.disc_length.seconds);

  cddb_make_request(server,hello,tmpbuffer,outbuffer);

  write(socket, outbuffer, strlen(outbuffer));
   
  cddb_skip_http(socket);
  cddb_read_line(socket, inbuffer, 256);

  strncpy(tmpbuffer, inbuffer, 256);
  switch(strtol(strtok(inbuffer, " "), NULL, 10)) {
    /* 200 - exact match */
  case 200:
	query->query_match = EXACT;
	query->query_matches = 1;
	strtok(tmpbuffer, "/");
	token=strtok(NULL,"/");

	if(token) 
	  strncpy(query->query_list[0].list_title,token+1, 64);
	else strcpy(query->query_list[0].list_title,"");

	strtok(tmpbuffer, " ");
	query->query_list[0].list_genre = cddb_genre_value(strtok(NULL, " "));
	strncpy(idtext, strtok(NULL, " "), 16);
	sscanf(idtext, "%xd", &query->query_list[0].list_id);
	strncpy(procbuffer, strtok(NULL, " "), 256);

	while((tmptr = strtok(NULL, " ")) != NULL)
	  sprintf(procbuffer,"%s %s", procbuffer, tmptr);

	if(procbuffer[strlen(procbuffer)-1]==0x0d)
	  procbuffer[strlen(procbuffer)-1]='\0';

	strncpy(query->query_list[0].list_artist, procbuffer, 64);
	
	break;
    /* 211 - inexact match */
  case 211:
	query->query_match = INEXACT;
	query->query_matches = 0;

	while(!cddb_read_line(socket, inbuffer, 256)) {
	  strtok(inbuffer, "/");
	  strncpy(query->query_list[query->query_matches].list_title,
			  strtok(NULL, "/") + 1, 64);
	  query->query_list[query->query_matches].list_genre =
		cddb_genre_value(strtok(inbuffer, " "));
	  strncpy(idtext, strtok(NULL, " "), 16);
	  sscanf(idtext, "%xd", &query->query_list[query->query_matches].list_id);
	 strncpy(procbuffer, strtok(NULL, " "), 256);

	 while((tmptr = strtok(NULL, " ")) != NULL)
	   snprintf(procbuffer, 256, "%s %s", procbuffer, tmptr);

	 strncpy(query->query_list[query->query_matches++].list_artist,
			 procbuffer, 64);
	}
      
	break;
    /* No match */
  default:
	query->query_match = NOMATCH;

	cddb_quit(socket);

	return -1;
  }

  cddb_quit(socket);
  
  return 0;
}

static int
cddb_data_mod(char *var, char *value, struct __unprocessed_disc_data *data)
{
  if(value == NULL)
	value = "";      
  
  if(strcmp(var, "DTITLE") == 0) {
	if(data->data_title_index >= MAX_EXTEMPORANEOUS_LINES)
	  return 0;
	strncpy(data->data_title[data->data_title_index++], value, 80);
  } else if(strncmp(var, "TTITLE", 6) == 0) {
	if(data->data_track[strtol((char *)var + 6, NULL, 10)].track_name_index
	   >= MAX_EXTEMPORANEOUS_LINES)
	   return 0;
	strncpy(data->data_track[strtol((char *)var + 6, NULL, 10)].
			track_name[data->data_track[strtol((char *)var + 6, NULL, 10)].
					  track_name_index++], value, 80);
  } else if(strcmp(var, "EXTD") == 0) {
	if(data->data_extended_index >= MAX_EXTENDED_LINES)
	  return 0;
	strncpy(data->data_extended[data->data_extended_index++], value, 80);
  } else if(strncmp(var, "EXTT", 4) == 0) {
	if(data->data_track[strtol((char *)var + 4, NULL, 10)].
	   track_extended_index >= MAX_EXTENDED_LINES)
	  return 0;
	strncpy(data->data_track[strtol((char *)var + 4, NULL, 10)].
			track_extended[data->data_track[strtol((char *)var+ 4,NULL,10)].
						  track_extended_index++], value, 80);
  }
  
  return 0;
}

static int
cddb_process_line(char *line, struct __unprocessed_disc_data *data)
{
  char *var, *value;
   
  if(strchr(line, '=') == NULL)
	return 0;
   
  line[strlen(line) - 1] = '\0';
  
  var = strtok(line, "=");
  if(var == NULL)
	return 0;

  value = strtok(NULL, "=");
  cddb_data_mod(var, value, data);
  
  return 0;
}

/* Read the actual CDDB entry */
int
cddb_read(int cd_desc,struct cddb_server *server,struct cddb_hello *hello,
		  struct cddb_entry *entry,struct disc_data *data)
{
  int socket;
  int index;
  char outbuffer[256], inbuffer[512],cmdbuffer[256];
  struct disc_info disc;
  struct __unprocessed_disc_data indata;
  
  socket=cddb_connect_server(server);
  if(socket==-1)
	return -1;
  
  cd_stat(cd_desc, &disc);
  
  indata.data_id = cddb_discid(cd_desc);   
  indata.data_genre = entry->entry_genre;
  indata.data_title_index = 0;
  indata.data_extended_index = 0;

  for(index = 0; index < disc.disc_totaltracks; index++) {
	indata.data_track[index].track_name_index = 0;
	indata.data_track[index].track_extended_index = 0;
  }

  snprintf(cmdbuffer,256,"cddb+read+%s+%08lx",cddb_genre(entry->entry_genre),
		   entry->entry_id);

  cddb_make_request(server,hello,cmdbuffer,outbuffer);

  write(socket, outbuffer, strlen(outbuffer));
   
  cddb_skip_http(socket);

  while(!cddb_read_line(socket, inbuffer, 512))
	cddb_process_line(inbuffer, &indata);

  data_format_input(data, &indata, disc.disc_totaltracks);

  cddb_quit(socket);
   
  return 0;
}

/* Process a single line in the sites list */
static int
cddb_sites_process_line(char *line, struct cddb_hostelement *host)
{
  char *ptr;
  
  strncpy(host->host_hostname, strtok(line, " "), 256);

  if(strcmp(strtok(NULL, " "),"http")) return -1;

  host->host_port = strtol(strtok(NULL, " "), NULL, 10);
  
  ptr = strtok(NULL, " ");
  host->host_icbm_latitude = strtol((char *)ptr + 1, NULL, 10);
  if(ptr[0] == 'N')
	host->host_icbm_latitude += 90;
  else
	host->host_icbm_latitude = 90 - host->host_icbm_latitude;
  
  ptr = strtok(NULL, " ");
  host->host_icbm_longitude = strtol((char *)ptr + 1, NULL, 10);
  if(ptr[0] == 'W')
	host->host_icbm_longitude += 180;
  else
	host->host_icbm_longitude = 180 - host->host_icbm_longitude;
  
  strcpy(host->host_city, "");
  
  while((ptr = strtok(NULL, " ")) != NULL)
	snprintf(host->host_city, 256, "%s %s", host->host_city, ptr);
  
  strcpy(host->host_city, (char *)host->host_city + 1);
  
  return 0;
}

/* Read the CDDB sites list */
static int
cddb_sites(struct cddb_server *server,struct cddb_serverlist *list)
{
  int socket;
  char outbuffer[256];
  char inbuffer[256];
  char hellobuf[256];
  struct cddb_hello hello;

  strncpy(hello.hello_program,PACKAGE,256);
  strncpy(hello.hello_version,VERSION,256);
  
  cddb_make_hello(&hello,hellobuf);

  socket=cddb_connect_server(server);
  if(socket==-1)
	return -1;

  snprintf(outbuffer,256,
		   "GET /~cddb/cddb.cgi?cmd=sites%s&proto=%s HTTP/1.0\n\n",
		   hellobuf,CDDB_PROTOCOL_LEVEL);

  write(socket,outbuffer,strlen(outbuffer));
  
  list->list_len = 0;
  
  cddb_skip_http(socket);

  while(!cddb_read_line(socket,inbuffer,256)) {
	if(cddb_sites_process_line(inbuffer,&list->list_host[list->list_len])!=-1)
	  list->list_len++;
  }

  cddb_quit(socket);
  
  return 0;
}

/* Bye! */
int
cddb_quit(int sock)
{
  char outbuffer[8];
   
  strcpy(outbuffer, "quit\n");
  write(sock, outbuffer, strlen(outbuffer));
  
  /* A bit nicer to the other side */
  shutdown(sock, 2);
  close(sock);
  
  return 0;
}

/* This is the function for completely automated CDDB operation */
int
cddb_read_data(int cd_desc, struct disc_data *data)
{
  int sock,index;
  struct disc_info disc;
  struct cddb_server server;
  struct cddb_entry entry;
  struct cddb_hello hello;
  struct cddb_query query;
  
  cd_stat(cd_desc, &disc);

  if(!disc.disc_present)
	return -1;
  
  server.server_index = 0;
  
  strncpy(hello.hello_program, PACKAGE, 256);
  strncpy(hello.hello_version, VERSION, 256);
  
  do {
	if(cddb_get_server(&server) < 0)
	  return -1;
	server.server_index++;
  }
  while(cddb_query(cd_desc,&server,&hello,&query)<0);
  
  /* Since this is an automated operation, we'll assume inexact matches are
	 correct. */
  
  entry.entry_id = query.query_list[0].list_id;
  entry.entry_genre = query.query_list[0].list_genre;
  
  if(cddb_read(cd_desc,&server,&hello,&entry,data) < 0)
	return -1;
  
  return 0;
}

/* Retrieve the server list */
int
cddb_read_sites(struct cddb_serverlist *list)
{
  struct cddb_server server;

  server.server_index = 0;
  
  do {
	if(cddb_get_server_force(&server) < 0)
	  return -1;

	server.server_index++;
  }
  while(cddb_sites(&server,list)<0);

  return 0;
}

/* Generate an entry for when CDDB is disabled/not working */
int
cddb_generate_unknown_entry(int cd_desc, struct disc_data *data)
{
  int index;
  struct disc_info disc;
  
  cd_stat(cd_desc, &disc);
  data->data_id = cddb_discid(cd_desc);
  strcpy(data->data_title, "Unknown");
  strcpy(data->data_artist, "Unknown");
  data->data_genre = CDDB_MISC;
  for(index = 0; index < disc.disc_totaltracks; index++) {
	strcpy(data->data_track[index].track_name, "Unknown");
  }
}

/* Generate an entry using CDDB if it's available, or Unknowns if it's not */
int
cddb_generate_new_entry(int cd_desc, struct disc_data *data)
{
  if(cddb_read_data(cd_desc, data) < 0) {
	cddb_generate_unknown_entry(cd_desc, data);
  }
  
  cddb_write_disc_data(cd_desc, data);
}

/* Read from the local database, using CDDB if there isn't an entry cached */
int
cddb_read_disc_data(int cd_desc, struct disc_data *outdata)
{
  FILE *cddb_data;
  int index;
  char root_dir[256], file[256], inbuffer[512];
  struct disc_info disc;
  struct stat st;
  struct __unprocessed_disc_data data;
  
  snprintf(root_dir, 256, "%s/.cddb", getenv("HOME"));
  
  if(stat(root_dir, &st) < 0) {
	if(errno != ENOENT)
	  return -1;
	else {
	  if(cddb_read_data(cd_desc, outdata) < 0)
		cddb_generate_unknown_entry(cd_desc, outdata);
	  cddb_write_disc_data(cd_desc, outdata);
	  return 0;
	}
  } else {
	if(!S_ISDIR(st.st_mode)) {
	  errno = ENOTDIR;
	  return -1;
	}
  }
  
  cd_stat(cd_desc, &disc);
  
  data.data_id = cddb_discid(cd_desc);
  data.data_title_index = 0;
  data.data_extended_index = 0;

  for(index = 0; index < disc.disc_totaltracks; index++) {
	data.data_track[index].track_name_index = 0;
	data.data_track[index].track_extended_index = 0;
  }
  
  for(index = 0; index < 12; index++) {
	snprintf(file, 256, "%s/%s/%08lx", root_dir, cddb_genre(index),
			 cddb_discid(cd_desc));
	
	if(stat(file, &st) == 0) {
	  cddb_data = fopen(file, "r");
	  
	  while(!feof(cddb_data)) {
	    fgets(inbuffer, 512, cddb_data);
	    cddb_process_line(inbuffer, &data);
	  }
	  
	  data.data_genre = index;
	  fclose(cddb_data);
	  
	  data_format_input(outdata, &data, disc.disc_totaltracks);
	  
	  return 0;
	}
  }
  
  if(cddb_read_data(cd_desc, outdata) < 0)
	cddb_generate_unknown_entry(cd_desc, outdata);
  
  cddb_write_disc_data(cd_desc, outdata);
  
  return 0;
}

/* Write to the local cache */
int
cddb_write_disc_data(int cd_desc, struct disc_data *indata)
{
  FILE *cddb_data;
  int index, tracks;
  char root_dir[256], genre_dir[256], file[256];
  struct stat st;
  struct disc_info disc;
  struct __unprocessed_disc_data data;
  
  cd_stat(cd_desc, &disc);
  data_format_output(&data, indata, disc.disc_totaltracks);
  
  snprintf(root_dir, 256, "%s/.cddb", getenv("HOME"));
  snprintf(genre_dir, 256, "%s/%s", root_dir, cddb_genre(data.data_genre));
  snprintf(file, 256, "%s/%08lx", genre_dir, data.data_id);
  
  if(stat(root_dir, &st) < 0) {
	if(errno != ENOENT)
	  return -1;
	else
	  mkdir(root_dir, 0755);
  } else {
	if(!S_ISDIR(st.st_mode)) {
	  errno = ENOTDIR;
	  return -1;
	}   
  }
  
  if(stat(genre_dir, &st) < 0) {
	if(errno != ENOENT)
	  return -1;
	else
	  mkdir(genre_dir, 0755);
  } else {
	if(!S_ISDIR(st.st_mode)) {
	  errno = ENOTDIR;
	  return -1;
	}
  }
  
  if((cddb_data = fopen(file, "w")) == NULL)
	return -1;
  
  fprintf(cddb_data, "# xmcd CD database file generated by %s %s\n",
		  PACKAGE, VERSION);
  fputs("# \n", cddb_data);
  fputs("# Track frame offsets:\n", cddb_data);

  for(index = 0; index < disc.disc_totaltracks; index++)
	fprintf(cddb_data, "#       %d\n", disc.track[index].track_start);

  fputs("# \n", cddb_data);
  fprintf(cddb_data, "# Disc length: %d seconds\n", disc.disc_length.minutes *
		  60 + disc.disc_length.seconds);
  fputs("# \n", cddb_data);
  fprintf(cddb_data, "# Revision: %d\n", CURRENT_CDDBREVISION);
  fprintf(cddb_data, "# Submitted via: %s %s\n", PACKAGE, VERSION);
  fputs("# \n", cddb_data);
  fprintf(cddb_data, "DISCID=%08lx\n", data.data_id);

  for(index = 0; index < data.data_title_index; index++)
	fprintf(cddb_data, "DTITLE=%s\n", data.data_title[index]);

  for(tracks = 0; tracks < disc.disc_totaltracks; tracks++) {
	for(index = 0; index < data.data_track[tracks].track_name_index; index++)
	  fprintf(cddb_data, "TTITLE%d=%s\n", tracks,
			  data.data_track[tracks].track_name[index]);
  }

  if(data.data_extended_index == 0)
	fputs("EXTD=\n", cddb_data);
  else {
	for(index = 0; index < data.data_extended_index; index++)
	  fprintf(cddb_data, "EXTD=%s\n", data.data_extended[index]);
  }
   
  for(tracks = 0; tracks < disc.disc_totaltracks; tracks++) {
	if(data.data_track[tracks].track_extended_index == 0)
	  fprintf(cddb_data, "EXTT%d=\n", tracks);
	else {
	  for(index = 0; index < data.data_track[tracks].track_extended_index;
		  index++)
		fprintf(cddb_data, "EXTT%d=%s\n", tracks,
				data.data_track[tracks].track_extended[index]);
	}
  }	 
  
  fputs("PLAYORDER=", cddb_data);
  
  fclose(cddb_data);
  
  return 0;
}

/* Delete an entry from the local cache based upon a data structure */
int
cddb_erase_entry(struct disc_data *data) {
  char root_dir[256], genre_dir[256], file[256];
  struct stat st;
  
  snprintf(root_dir, 256, "%s/.cddb", getenv("HOME"));
  snprintf(genre_dir, 256, "%s/%s", root_dir, cddb_genre(data->data_genre));
  snprintf(file, 256, "%s/%08lx", genre_dir, data->data_id);
  
  if(stat(root_dir, &st) < 0) {
	if(errno != ENOENT)
	  return -1;
	else
	  return 0;
  } else {
	if(!S_ISDIR(st.st_mode))
	  return 0;  
  }
  
  if(stat(genre_dir, &st) < 0) {
	if(errno != ENOENT)
	  return -1;
	else
	  return 0;
  } else {
	if(!S_ISDIR(st.st_mode))
	  return 0;
  }
  
  if(unlink(file) < 0) {
	if(errno != ENOENT)
	  return -1;
  }
  
  return 0;
}

int
cddb_stat_disc_data(int cd_desc, struct cddb_entry *entry)
{
  int index;
  struct disc_info disc;
  struct stat st;
  char root_dir[256], file[256], inbuffer[512];
  
  cd_stat(cd_desc, &disc);
  entry->entry_id = cddb_discid(cd_desc);
  
  snprintf(root_dir, 256, "%s/.cddb", getenv("HOME"));
  
  if(stat(root_dir, &st) < 0) {
	if(errno != ENOENT)
	  return -1;
	else {
	  entry->entry_present = 0;
	  return 0;
	}
  } else {
	if(!S_ISDIR(st.st_mode)) {
	  errno = ENOTDIR;
	  return -1;
	}
  }
  
  for(index = 0; index < 12; index++) {
	snprintf(file, 256, "%s/%s/%08lx", root_dir, cddb_genre(index),
			 cddb_discid(cd_desc));

	if(stat(file, &st) == 0) {
	  entry->entry_present = 1;
	  entry->entry_genre = index;
	  
	  return 0;
	}
  }
   
  entry->entry_present = 0;
  
  return 0;
}
